define([
    'underscore',
    'backbone'
  ], function(
    _,
    Backbone
  ) {
  'use strict';

  var remoteAPIVersion, localAPIVersion, remoteTime, localTime;

  /**
   * Method to handle the XMLHttpRequest.
   * @method XHR
   * @access public
   * @param object pack
   * @param info text to happend to error detail in case of failure
   */
  var XHR = function (pack)  {
    var xhr = null;
    pack.info = pack.info || '';

    if(!pack.hasOwnProperty("async")) {
      pack.async = true;
    }

    // ------------ request TYPE
    pack.type = pack.type || 'POST';
    // method override test
    if (false && pack.type === 'DELETE') {
      pack.type = 'POST';
      pack.params =  pack.params || {};
      pack.params['_method'] = 'DELETE';
    }

    // ------------- request HEADERS


    pack.headers = pack.headers || {};

    if (pack.type === 'POST' || pack.type === 'PUT') {
      if (pack.headers['Content-Type'] &&
          pack.headers['Content-Type'] === 'multipart/form-data') {
          /* Delete the Content-Type to let browser auto-fill
           * it along with the required data boundary. */
          delete pack.headers['Content-Type'];
      } else {
        /* add json headers by default if POST or PUT. */
        pack.headers['Content-Type'] =
          pack.headers['Content-Type'] || 'application/json; charset=utf-8';
      }
    }

    if (pack.type === 'POST') { pack.params = pack.params || {}; }

    // -------------- error
    pack.error = pack.error || function (error, context) {
      throw new Error(JSON.stringify(error, function (key, value) {
        if (value === null) { return; }
        if (value === '') { return; }
        return value;
      }, 2));
    };

    var detail = pack.info + ', req: ' + pack.type + ' ' + pack.url;

    // --------------- request
    xhr = _initXHR();

    xhr.open(pack.type, pack.url, pack.async);
    xhr.withCredentials = true;


    xhr.onreadystatechange = function () {
      detail += ' xhrstatus:' + xhr.statusText;
      if (xhr.readyState === 0) {
        pack.error({message: "pryvXHRCall unsent", detail: detail, id: "INTERNAL_ERROR", xhr: xhr}, pack.context);
      } else if (xhr.readyState === 4) {
        var result = null;

        try { result = JSON.parse(xhr.responseText); } catch (e) {
          return pack.error({message: "Data is not JSON", detail: xhr.responseText+"\n"+detail, id: "RESULT_NOT_JSON", xhr: xhr}, pack.context);
        };

        pack.success(result,pack.context,xhr);
      }
    };

    for(var key in pack.headers){
      if (pack.headers.hasOwnProperty(key))
        xhr.setRequestHeader(key, pack.headers[key]);
    }

    //--- prepare the params
    var sentParams = null;
    if (pack.params) {
      if (pack.headers['Content-Type']) {
        /* json Content-Type: try to stringify the data. */
        try {
          sentParams = JSON.stringify(pack.params)
        } catch (e) {
          return pack.error({message: "Parameters are not JSON", detail: "params: "+pack.params+"\n "+detail, id: "INTERNAL_ERROR", error: e}, pack.context);
        };
      } else {
        /* It's a formData. Send it without stringifying it.
         * The browser will then automatically fill the Content-Type. */
        sentParams = pack.params;
      }
    }

    //--- sending the request
    try {
      xhr.send(sentParams);
    } catch (e) {
      return pack.error({message: "pryvXHRCall unsent", detail: detail, id: "INTERNAL_ERROR", error: e}, pack.context);
    };
    return xhr;
  };

  /**
   * Method to initialize XMLHttpRequest.
   * @method _initXHR
   * @access private
   * @return object
   */
  var _initXHR = function() {
    var XHR = null;

    try { XHR = new XMLHttpRequest(); }
    catch(e) {
      try { XHR = new ActiveXObject("Msxml2.XMLHTTP"); }
      catch (e2) {
        try { XHR = new ActiveXObject("Microsoft.XMLHTTP"); }
        catch (e3) { ;
          console.log('XMLHttpRequest implementation not found.');
        };
        console.log('XMLHttpRequest implementation not found.');
      };
      console.log('XMLHttpRequest implementation not found.');
    }
    return XHR;
  };

  /* Backbone hacking to get API-Version and
   * Server-Time headers. */
  Backbone._ajax = Backbone.ajax;
  Backbone.ajax = function(){
    Backbone._ajax.apply(this, arguments)
    .success(function(res, textStatus, jqXHR){
      localTime = new Date().getTime();
      /* Check API version. */
      remoteAPIVersion = jqXHR.getResponseHeader('API-Version');
      if (remoteAPIVersion) {
        if (localAPIVersion) {
          if (localAPIVersion !== remoteAPIVersion) {
            /* API version has changed! */
            console.log('WARNING: API version updated to ' + remoteAPIVersion);
          }
        } else {
          localAPIVersion = remoteAPIVersion;
        }
      }

      /* Save server time. */
      remoteTime = jqXHR.getResponseHeader('Server-Time');
    });
  };

  return {
    getServerVersion: function(){
      return localAPIVersion;
    },
    localToRemoteTime: function(local){
      var delta = this.getDeltaTime(),
          remoteMili = local + delta;
      return remoteMili / 1000;
    },
    remoteToLocalTime: function(remote){
      var delta = this.getDeltaTime(),
          remoteMili = remote*1000;
      return remoteMili - delta;
    },
    getDeltaTime: function(){
      if (remoteTime && localTime) {
        return (remoteTime*1000 - localTime);
      }
      return 0;
    },
    whoAmI: function(data){
      XHR({
        url: data.url + '/admin/who-am-i',
        type: 'GET',
        success: function (result, context, xhr) {
          data.success(result, context, xhr);
        },
        error: function(response, context){
          data.error(response, context);
        }
      });
    },
    post:function(pack){
      XHR(pack);
    },

    /**
     * To upload one or more files, add the required
     * headers in pack.
     */
    postFiles : function(pack) {
      /* Add required headers for sending formData. */
      pack.headers = {
        'Content-Type': 'multipart/form-data',
        'X-Requested-With': 'XMLHttpRequest'
      };
      XHR(pack);
    }
  };
});
